'use strict';

/* For ESLint: List any global identifiers used in this file below */
/* global log, Filter, SelectorFilter, ElemHideEmulationFilter, PatternFilter,
ContentFilter,
ElementTypes, SAFARI, DomainSet, getExcludeFiltersText, getUnicodeDomain */
/* eslint-disable no-bitwise */

// Converts non-standard filters to a standard format, and removes
// invalid filters.
const FilterNormalizer = {

  userExcludedFilterArray: [],

  setExcludeFilters(text) {
    if (text) {
      FilterNormalizer.userExcludedFilterArray = text.split('\n');
    } else {
      FilterNormalizer.userExcludedFilterArray = null;
    }
  },

  // Normalize a set of filters.
  // Remove broken filters, useless comments and unsupported things.
  // Input: text:string filter strings separated by '\n'
  //        keepComments:boolean if true, comments will not be removed
  //        allowSnippets:boolean if true, Snippets filters will be allowed
  // Returns: filter strings separated by '\n' with invalid filters
  //          removed or modified
  normalizeList(text, keepComments, allowSnippets = false) {
    const result = [];
    let ignoredFilterCount = 0;
    if (text) {
      const lines = text.split('\n');
      for (let i = 0; i < lines.length; i += 1) {
        try {
          const newfilter = FilterNormalizer.normalizeLine(lines[i], allowSnippets);
          if (newfilter) {
            result.push(newfilter);
          } else if (newfilter !== false) {
            ignoredFilterCount += 1;
          } else if (keepComments) {
            result.push(lines[i]);
          }
        } catch (ex) {
          log(`Filter '${lines[i]}' could not be parsed: ${ex}`);
          ignoredFilterCount += 1;
        }
      }

      if (ignoredFilterCount) {
        log(`Ignoring ${ignoredFilterCount} rule(s)`);
      }
    }
    return `${result.join('\n')}\n`;
  },

  // Normalize a single filter.
  // Input: filter:string a single filter
  //        allowSnippets:boolean if true, Snippets filters will be allowed
  // Return: normalized filter string if the filter is valid, null if the filter
  //         will be ignored or false if it isn't supposed to be a filter.
  // Throws: exception if filter could not be parsed.
  //
  // Note that 'Expires' comments are considered valid comments that
  // need retention, because they carry information.
  normalizeLine(filter, allowSnippets = false) {
    // Some rules are separated by \r\n; and hey, some rules may
    // have leading or trailing whitespace for some reason.
    const trimmedFilter = filter.replace(/\r$/, '').trim();

    // Remove comment/empty filters.
    if (Filter.isComment(trimmedFilter)) {
      return false;
    }
    let cleanedFilter = FilterNormalizer.cleanFilterText(trimmedFilter);

    // Convert old-style hiding rules to new-style.
    if (/#[*a-z0-9_-]*(\(|$)/.test(cleanedFilter)
        && !/#@?#./.test(cleanedFilter)
        && !Filter.isContentFilter(cleanedFilter)
        && !Filter.isAdvancedSelectorFilter(cleanedFilter)) {
      // Throws exception if unparseable.
      const oldFilter = cleanedFilter;
      cleanedFilter = FilterNormalizer.oldStyleHidingToNew(cleanedFilter);
      log(`Converted ${oldFilter} to ${cleanedFilter}`);
    }
    const match = Filter.elemhideRegExp.exec(cleanedFilter);
    if (match) {
      let propsMatch;
      if (!match[2]
          && (propsMatch = Filter.oldAdvanceSyntaxRegExp.exec(match[3]))) {
        const oldFilter = cleanedFilter;
        // This is legacy CSS properties syntax, convert to current syntax
        const prefix = match[3].substr(0, propsMatch.index);
        const expression = propsMatch[2];
        const suffix = match[3].substr(propsMatch.index + propsMatch[0].length);
        cleanedFilter = `${match[1]}#?#${prefix}:-abp-properties(${expression})${suffix}`;
        log(`Converted ${oldFilter} to ${cleanedFilter}`);
      }
    }

    if (typeof FilterNormalizer.userExcludedFilterArray !== 'undefined'
        && FilterNormalizer.userExcludedFilterArray
        && FilterNormalizer.userExcludedFilterArray.length > 0
        && FilterNormalizer.userExcludedFilterArray.indexOf(cleanedFilter) >= 0) {
      return null;
    }

    // If it is a hiding rule...
    let parsedFilter;
    if (Filter.isSelectorFilter(cleanedFilter)) {
      // The filter must be of a correct syntax
      const selectorPart = cleanedFilter.replace(/^.*?#@?\??#/, '');
      try {
        // Throws if the filter is invalid...
        if (document.querySelector(`${selectorPart},html`).length === 0) {
          throw new Error('Causes other filters to fail');
        }
      } catch (ex) {
        // ...however, the thing it throws is not human-readable. This is.
        throw new Error(`Invalid CSS selector syntax: ${selectorPart}`);
      }

      // On a few sites, we have to ignore [style] rules.
      // Affects Chrome (crbug 68705) and Safari (issue 6225).
      // Don't exclude the sites unless the filter would apply to them, or
      // loading the site will hang in Safari 6 while Safari creates a bunch of
      // one-off style sheets (issue 7356).
      if (/style([\^$*]?=|\])/.test(cleanedFilter)) {
        const excludedDomains = ['mail.google.com', 'mail.yahoo.com'];
        cleanedFilter = FilterNormalizer.ensureExcluded(cleanedFilter, excludedDomains);
      }

      parsedFilter = new SelectorFilter(cleanedFilter);
    } else if (Filter.isAdvancedSelectorFilter(cleanedFilter)) {
      parsedFilter = new ElemHideEmulationFilter(cleanedFilter);
    } else if (Filter.isContentFilter(cleanedFilter)) {
      if (Filter.isSnippetFilter(cleanedFilter) && allowSnippets === false) {
        throw new Error('Snippet rule is not allowed'); // snippets should be ignored
      }
      parsedFilter = ContentFilter.fromText(cleanedFilter);
    } else { // If it is a blocking rule...
      parsedFilter = PatternFilter.fromText(cleanedFilter); // throws if invalid
      const types = parsedFilter.allowedElementTypes;

      const whitelistOptions = (ElementTypes.document | ElementTypes.elemhide);
      const hasWhitelistOptions = types & whitelistOptions;
      if (!Filter.isWhitelistFilter(cleanedFilter) && hasWhitelistOptions) {
        throw new Error('$document and $elemhide may only be used on whitelist filters');
      }

      // We are ignoring Hulu whitelist filter, so user won't see ads in videos
      // but just a message about using AdBlock - Issue 7178
      // We are also ignoring Google whitelist filter to prevent whitelisting some ads
      // e.g. on YouTube by Danish filter list - Issue #264
      if (!SAFARI
          && (/^@@\|\|hulu\.com\/published\/\*\.(flv|mp4)$/.test(cleanedFilter)
            || /^@@googleads.g.doubleclick.net/.test(cleanedFilter))) {
        return null; // Hulu-only: background.js implements this rule more specifically
      }
    }

    // Ignore filters whose domains aren't formatted properly.
    FilterNormalizer.verifyDomains(parsedFilter.domains);

    // Ensure filter doesn't break AdBlock
    FilterNormalizer.checkForObjectProperty(cleanedFilter);

    // Nothing's wrong with the filter.
    return cleanedFilter;
  },
  /**
   * Removes unnecessary whitespaces from filter text, will only return null if
   * the input parameter is null.
   * @param {string} text
   * @return {string}
   */
  cleanFilterText(textArg) {
    if (!textArg) {
      return textArg;
    }
    let text = textArg;

    // Remove line breaks, tabs etc
    text = text.replace(/[^\S ]+/g, '');

    // Don't remove spaces inside comments
    if (/^ *!/.test(text)) {
      return text.trim();
    }

    // Special treatment for content filters, right side is allowed to contain
    // spaces
    if (Filter.contentRegExp.test(text)) {
      const [, domains, separator, body] = /^(.*?)(#[@?$]?#?)(.*)$/.exec(text);
      return domains.replace(/ +/g, '') + separator + body.trim();
    }

    // For most regexp filters we strip all spaces, but $csp filter options
    // are allowed to contain single (non trailing) spaces.
    const strippedText = text.replace(/ +/g, '');
    if (!strippedText.includes('$') || !/\bcsp=/i.test(strippedText)) {
      return strippedText;
    }

    const optionsMatch = Filter.optionsRegex.exec(strippedText);
    if (!optionsMatch) {
      return strippedText;
    }

    // For $csp filters we must first separate out the options part of the
    // text, being careful to preserve its spaces.
    const beforeOptions = strippedText.substring(0, optionsMatch.index);
    let strippedDollarIndex = -1;
    let dollarIndex = -1;
    do {
      strippedDollarIndex = beforeOptions.indexOf('$', strippedDollarIndex + 1);
      dollarIndex = text.indexOf('$', dollarIndex + 1);
    }
    while (strippedDollarIndex !== -1);
    const optionsText = text.substr(dollarIndex + 1);

    // Then we can normalize spaces in the options part safely
    const options = optionsText.split(',');
    for (let i = 0; i < options.length; i += 1) {
      const option = options[i];
      const cspMatch = /^ *c *s *p *=/i.exec(option);
      if (cspMatch) {
        options[i] = cspMatch[0].replace(/ +/g, '')
                     + option.substr(cspMatch[0].length).trim().replace(/ +/g, ' ');
      } else {
        options[i] = option.replace(/ +/g, '');
      }
    }

    return `${beforeOptions}$${options.join()}`;
  },
  // validates a single filter
  // used by option pages, content scripts, etc
  validateLine(filter, returnException, allowSnippets = false) {
    try {
      return FilterNormalizer.normalizeLine(filter, allowSnippets);
    } catch (ex) {
      if (returnException) {
        return ({ exception: ex.message });
      }
      return false;
    }
  },

  // validates an array of filters
  // used by option pages, content scripts, etc
  validateList(filters, returnException, allowSnippets = false) {
    let filterText;
    try {
      for (let i = 0; i < filters.length; i += 1) {
        filterText = filters[i];
        FilterNormalizer.normalizeLine(filterText, allowSnippets);
      }
    } catch (ex) {
      if (returnException) {
        return ({ exception: ex.message, filter: filterText });
      }
      return false;
    }

    return true;
  },

  // Return |selectorFilterText| modified if necessary so that it applies to no
  // domain in the |excludedDomains| list.
  // Throws if |selectorFilterText| is not a valid filter.
  // Example: ("a.com##div", ["sub.a.com", "b.com"]) -> "a.com,~sub.a.com##div"
  ensureExcluded(selectorFilterText, excludedDomains) {
    let text = selectorFilterText;
    const filter = new SelectorFilter(text);
    const mustExclude = excludedDomains.filter(domain => filter.domains.computedHas(domain));

    if (mustExclude.length > 0) {
      let toPrepend = `~${mustExclude.join(',~')}`;
      if (text[0] !== '#') {
        toPrepend += ',';
      }
      text = toPrepend + text;
    }

    return text;
  },

  // Convert an old-style hiding rule to a new one.
  // Input: filter:string old-style filter
  // Returns: string new-style filter
  // Throws: exception if filter is unparseable.
  oldStyleHidingToNew(filter) {
    // Old-style is domain#node(attr=value) or domain#node(attr)
    // domain and node are optional, and there can be many () parts.
    const newFilter = filter.replace('#', '##');
    const parts = newFilter.split('##'); // -> [domain, rule]
    const domain = parts[0];
    const rule = parts[1];

    // Make sure the rule has only the following two things:
    // 1. a node -- this is optional and must be '*' or alphanumeric
    // 2. a series of ()-delimited arbitrary strings -- also optional
    //    the ()s can't be empty, and can't start with '='
    if (rule.length === 0
        || !/^(?:\*|[a-z0-9\-_]*)(?:\([^=][^)]*?\))*$/i.test(rule)) {
      throw new Error('bad selector filter');
    }

    const firstSegment = rule.indexOf('(');

    if (firstSegment === -1) {
      return `${domain}##${rule}`;
    }

    const node = rule.substring(0, firstSegment);
    let segments = rule.substring(firstSegment);

    // turn all (foo) groups into [foo]
    segments = segments.replace(/\((.*?)\)/g, '[$1]');

    // turn all [foo=bar baz] groups into [foo="bar baz"]
    // Specifically match:    = then not " then anything till ]
    segments = segments.replace(/=([^"][^\]]*)/g, '="$1"');

    // turn all [foo] into .foo, #foo
    // #div(adblock) means all divs with class or id adblock
    // class must be a single class, not multiple (not #*(ad listitem))
    // I haven't ever seen filters like #div(foo)(anotherfoo), so ignore these
    let resultFilter = node + segments;
    const match = resultFilter.match(/\[([^=]*?)\]/);
    if (match) {
      resultFilter = `${resultFilter.replace(match[0], `#${match[1]}`)
      },${resultFilter.replace(match[0], `.${match[1]}`)}`;
    }

    return `${domain}##${resultFilter}`;
  },

  // Checks if the filter is an object property, which we should not overwrite.
  // See Issue 7117.
  // Throw an exeption if that's the case
  // Input: text (string): the item to check
  checkForObjectProperty(text) {
    if (text in Object) {
      throw new Error('Filter causes problems in the code');
    }
  },

  // Throw an exception if the DomainSet |domainSet| contains invalid domains.
  verifyDomains(domainSet) {
    for (let domain in domainSet.has) {
      if (domain !== DomainSet.ALL) {
        // Convert punycode domains to Unicode
        domain = getUnicodeDomain(domain);
        if (/^([%a-z0-9\-_\u00DF-\u00F6\u00F8-\uFFFFFF]+\.)*[%a-z0-9\u00DF-\u00F6\u00F8-\uFFFFFF]+\.?$/i.test(domain) === false) {
          throw new Error(`Invalid domain: ${domain}`);
        }

        // Ensure domain doesn't break AdBlock
        FilterNormalizer.checkForObjectProperty(domain);
      }
    }
  },
};

// Initialize the exclude filters at startup
try {
  getExcludeFiltersText((excludeFilters) => {
    FilterNormalizer.setExcludeFilters(excludeFilters);
  });
} catch (e) {

  // ignore exception in Safari on options / resource block pages
}

// Attach object to window
Object.assign(window, {
  FilterNormalizer,
});
